;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :stella)

(defvar *csound-command-string*
  (namestring (merge-pathnames "playcsound" *cm-site-directory*)))

(defvar *soundfile-play-command-string*
  (namestring (merge-pathnames "playsnd" *cm-site-directory*)))

(defsyntax csound
  :pathname (pathname "test.sco")
  :stream-types '((csound-score-file "sco"))
  :nicknames '(csnd))

(defobject csound-score-file (event-file header-mixin)
  ((orchestra :initarg orchestra :initform nil)
   (output :initarg output :initform nil)
   (syntax :initform (find-syntax ':csound))
   (args :initarg args :initform nil))
  (:parameters start end timescale orchestra output args header))

(defmethod initialize-stream-for-processing ((stream csound-score-file))
  (let ((file (slot-value stream 'stream)))
    (format file "; Common Music output of ~A~%~%" (date-string))
    (when (slot-boundp stream 'header)
      (format file "~A~%" (slot-value stream 'header)))
    (format file "~&s~%")))

(defmethod deinitialize-stream-for-processing ((stream csound-score-file))
  (let ((file (slot-value stream 'stream)))
    (format file "e~%")))

;;;
;;; Output specification may occur in different places: (1) the "output" slot,
;;; (2) passed via the "args" option on the Stella command line, and (3)
;;; the "args" slot.  The order of precedence is then: (1) -- (3) -- (2).

(defmethod play-using-syntax ((syntax csound) file &rest arglist)
  (declare (ignore #+mcl file arglist))
  #+mcl
  (format t "Sorry, can't play csound from Lisp on the Mac.~%")
  #-mcl
  (let* ((whitespace '(#\space #\tab #\newline))
         (orc (slot-value (find-stream file) 'orchestra))
         ;(command-options (or (cadr (member 'args arglist))))
         ;; args slot already holds command-options at this time, if any
         (options (slot-value (find-stream file) 'args))
         ;(options (or command-options stream-options))
         (ofile (slot-value (find-stream file) 'output))
         (o-flag (search "-o" options))
         (o-flag-spec (when o-flag
                        (let ((left (position-if
                                     #'(lambda (c)
                                         (not (member c whitespace)))
                                     options :start (+ o-flag 2))))
                          (when left
                            (subseq options left
                                    (position-if #'(lambda (c)
                                                     (member c whitespace))
                                                 options :start left))))))
         (out (or o-flag-spec ofile))
         (play? (not (string= out "dac"))))
    
    (cm::shell
     (format nil "~A ~:[~;~:*~a ~] ~:[-o ~a ~;~*~]~A ~A"
             ;; we don't want out if -o option has been passed
             *csound-command-string* options
             o-flag out orc (namestring (truename file))))
    
    (when play?
      (tell-user "Playing ~A ." out)
      (cm::shell (format nil "~A ~A"
                         *soundfile-play-command-string* out)))))


;;;
;;; csound-note implements automatic score printing for all subclasses.
;;;

(defclass csound-note (note)
  ((ins :initarg ins :initarg :ins)))

(defmethod print-object ((object csound-note) stream)
  (let ((*print-case* ':downcase))
    (write-string "#<" stream)
    (loop with d
          for p in (object-parameters object)
          for s = (parameter-name p)
          do
          (if d (write-char d stream) (setf d #\Space))
          (princ (slot-value-or-default
                  object (if (eq s 'time) 'rhythm s) +slot-unset+)
                 stream)
          (setf d #\Space))
    (write-char #\> stream)))

(defmethod initialize-instance :after ((object csound-note) &rest args)
  (declare (ignore args))
  (unless (slot-boundp object 'ins)
    (setf (slot-value object 'ins)
          (class-name (class-of object)))))

(defmethod write-event ((object csound-note) (stream csound-score-file))
  (let ((*print-case* ':downcase))
    (loop with f = (slot-value stream 'stream)
          and d
          for p in (object-parameters object)
          do
          (if d (write-char d f) (setf d #\Space))
          (let ((v (slot-value object (parameter-name p))))
            (princ (if (floatp v)
                     (locally (declare (optimize (speed 3) (safety 0)))
                       (/ (the fixnum (floor (the float v) .001))
                          1000.0))
                     v)
                   f))
          finally (terpri f))))

(defmethod import-using-syntax ((syntax csound) file &rest args)
  (apply #'csound-import file (canonicalize-pairs args nil nil ':keyword)))

(defun csound-map-scofile (file fun start end)
  ;; strip header by reading until s.  is this right?
  (loop with quit
        for line = (read-line file nil ':eof)
        until (eq line ':eof)
        unless (string= line "")
        do (setf quit (char= (elt line 0) #\s) )
        until quit
        finally (unless quit
                  (warn-user "Didn't find \"s\" marker line.")
                  (return-from csound-map-scofile nil)))
  (loop with quit and char
        for line = (read-line file nil ':eof)
        do
        (cond ((eq line ':eof) (setf quit t))
              ((char= (setf char (elt line 0)) #\e) (setf quit t))
              ((char= char #\i)
               (let ((list (string-forms line)))
                 (when (or (not start) (>= (second list) start))
                   (if (or (not end) (<= (second list) end))
                       (funcall fun list)
                       (setf quit t))))))
        until quit)
   t)

(defun csound-import (scofile &key start end container insmap)
  (let ((list '())
        (ins (make-hash-table)) )
    (flet ((setins (i c &aux v)
             (if (and (setf v (find-class c))
                      (typep (class-prototype v) 'csound-note))
               (setf (gethash i ins)
                     ; cache: (#<class> "ins1" . parameters)
                     (list* v (string-downcase (symbol-name c))
                            (object-parameters v)))
               (error "No Csound note class for instrument ~S." c))))
      ;; parse user supplied instrument map.
      (when insmap
        (loop for (i c) on insmap by #'cddr do (setins i c)))
      (with-open-file (file scofile :direction :input)
        (flet ((fn (l)
                 (let* ((i (pop l))
                        (d (or (gethash i ins) (setins i i)))
                        (n (make-instance (first d) :ins (second d))))
                   (loop for v in l         ; ins name popped
                         for p in (cdddr d) ; skip class, ins & ins par
                         do (setf (slot-value n (parameter-name p)) v))
                   (when list
                     (setf (object-rhythm (car list))
                           (- (first l) (object-time (first list)))))
                   (push n list))))
          (csound-map-scofile file #'fn start end))
        ;; Last object parsed has no rhythm, try setting it to a
        ;; duration value otherwise 0.
        (when list
          (setf (object-rhythm (car list))
                (cond ((slot-exists-p (car list) 'duration)
                       (slot-value (car list) 'duration))
                      ((slot-exists-p (car list) 'dur)
                       (slot-value (car list) 'dur))
                      (t 0.0)))
          ;; find a new name for the conainer
          (unless container
            (let ((root (format nil "FROM-~:@(~A~)"
                                (pathname-name file))))
              (loop with name = root and num = 0
                    for sym = (find-symbol name)
                    while (and sym (find-object sym))
                    do (setf name (format nil "~A-~D" root (incf num)))
                    finally
                    (setf container
                          (make-instance 'thread :id (intern name))))))
          (remove-all-objects container)
          (add-objects (nreverse list) container nil :copy-first nil))
        container))))


;;;
;;; -*- EOF -*-

