;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :common-music)

#-(or excl kcl mcl aclpc lispworks clisp cmu)
(eval-when (load compile eval) (error "Fix impl.lisp for this port."))

;;;
;;; evey os/lisp combination must define their file types.
;;;

(defconstant +source-type+ #-(or dos os/2 win32) "lisp"
                           #+(or dos os/2 win32) "LSP")
(defconstant +binary-type+  
  #+excl "fasl"
  #+mcl (pathname-type ccl:*.fasl-pathname*)
  #+kcl "o"
  #+aclpc "fsl"
  #+(and lispworks mips) "mfasl"
  #+(and clisp (or dos os/2 win32)) "FAS"
  #+(and clisp (not dos) (not os/2) 
         (not win32)) "fas"
  #+cmu (c:backend-fasl-file-type c:*backend*)
  )
(defconstant +midi-type+   #-dos "midi" #+dos "MID")
(defconstant +text-type+   #-dos "text" #+dos "TXT")
(defconstant +cm-type+ #-dos "cm" #+dos "CM") 
(defconstant +sound-type+ #+mac "aiff" #+dos "SND" #-(or mac dos) "snd")

#+unix
(defun unix-image-directory (name path)
  (if (or (char= (elt name 0) #\/)
          (char= (elt name 0) #\.))
    (let ((pos (position #\/ name :from-end t)))
      (if (not pos) "./" (subseq name 0 (1+ pos))))
    (loop with beg = 0 and len = (length path) 
          while (< beg len)
          for end = (or (position #\: path :start beg) len)
          for dir = (subseq path beg end)
          do (unless (char= (elt dir (1- (length dir))) #\/)
               (setf dir (concatenate 'string dir "/")))
          when (probe-file (make-pathname :defaults (pathname dir) 
                                          :name name))
          return dir
          do (setf beg (1+ end )))))


(defun load-cminit (&aux (name #+DOS "CMINIT" #-DOS "cminit"))
  ;; load cminit.lisp if one can be found.
  (flet ((maybe-load (dir)
           (let ((fasl (make-pathname :defaults dir :name name
                                      :type +binary-type+))
		 (lisp (make-pathname :defaults dir :name name 
                                      :type +source-type+)))
             (if (probe-file fasl)
               (load fasl :verbose nil)
               (if (probe-file lisp)
                 (load lisp :verbose nil)
                 nil)))))
    (or (maybe-load *cm-site-directory*)
        (maybe-load *common-music-directory*)
        (maybe-load (image-directory)))))

;;;
;;; Every port must define (or stub out) the following functions.
;;;
;;; without-interrupts (&body body)          Execute body with priority.         
;;; shell (string)                           Invoke Unix with command
;;; image-directory ()                       Directory of booted image.
;;; save-cm (path)                           Save CM image file.
;;;
;;; If the Lisp implementation supports multiprocessing, you should
;;; override the default definitions for these:
;;;
;;; process-run-function                     Run function in new process
;;; process-wait                             Wait until fn returns true
;;;

#-(or processes multiprocessing)            ; use one process.
(progn
  (defun process-run-function (options fn args)
    (declare (ignore options))
    (apply fn args))
  (defun process-wait (reason fn &rest args)
    (declare (ignore reason))
    (loop until (apply fn args)))
  )

#+mcl
(progn
  ; moved to make-cm.lisp
  ;(eval-when (:compile-toplevel :execute :load-toplevel) 
  ;  (ccl:require :resources))
  
  #+processes
  (import '(ccl:process-run-function ccl:process-wait))
  
  (defun address-of (thing) (ccl:%address-of thing))
  
  (defmacro without-interrupts (&body body) 
    `(ccl:without-interrupts ,@body))
  
  (defun shell (string) (warn "Shell not implemented in MCL: ~S" string))
  
  
  (defun image-directory () (translate-logical-pathname "HOME:"))
  
  (defvar *cm-resources* nil)
  
  (defun save-cm (pathname)
    (setf ccl:*lisp-startup-functions*
          (append ccl:*lisp-startup-functions*
                  (list #'(lambda ()
                            (ccl:egc t)
                            (setf *readtable* *common-music-readtable*)
                            (setf *package* (find-package :cm))
                            (load-cminit)))))
    (let ((resfile (merge-pathnames ":capella:mac:cm.rsrc"
                                    *common-music-directory*)))
      (ccl:with-open-resource-file (ref resfile)
        (ccl:using-resource-file ref
          (loop for id from #x4000 to #x400c    ; id's hardcoded
                append
                (loop for type in '(:|ics\#| :|ics4| :|ics8|)
                      for res
                      = (list 
                         (let ((r (ccl:require-trap _getresource type id)))
                           (ccl:require-trap _loadresource r)
                           (ccl:require-trap _detachresource r)
                           r) type id)
                      when res
                      do (pushnew res *cm-resources*)
                      else do
                      (warn "Resource ~a ID ~d not found in ~a"
                            (symbol-name :|ics\#|) id resfile)))))
      
      (ccl:save-application (namestring pathname)
                            :size '(#-clm #xA00000 #+clm #xC00000 #x500000)
                            ; :excise-compiler t
                            :init-file nil :resources *cm-resources*)))
  
  )

#+excl
(progn

  #+multiprocessing
  (import '(mp:process-run-function mp:process-wait))

  (defun address-of (thing) (excl::pointer-to-fixnum thing))
 
  (defmacro without-interrupts (&body body) 
    `(excl:without-interrupts ,@body))
 
  (defun shell (string) (excl:shell string))

  (defun image-directory ()
    (unix-image-directory (system:command-line-argument 0)
                          (system:getenv "PATH")))  
  
  (defun save-cm (pathname &aux fn)
    (setf fn #'(lambda ()
                 (tpl:setq-default *readtable*
                                   cm::*common-music-readtable*)
                 (tpl:setq-default *package* (find-package :cm))
                 (cm::load-cminit)))
    #-allegro-v4.3
    (excl:dumplisp :name (namestring pathname) :checkpoint nil
                   :restart-function fn :read-init-file t)
    #+allegro-v4.3
    (progn
      (setf excl::*read-init-files* t
	    ;;excl:*restart-init-function* fn
	    )
      (push fn excl::*restart-actions*)
      (excl:dumplisp :name (namestring pathname) :checkpoint nil)))
  )

#+akcl
(progn
  
  (defun address-of (thing) (si:address thing))

  (defmacro without-interrupts (&body body) `(progn ,@body))
  
  (defun shell (string) (lisp:system string)) 
 
  (defun image-directory () 
    (truename 
     (make-pathname
      :directory (pathname-directory si::*system-directory*))))

  (defun save-cm (pathname)
    (let ((*standard-input* *terminal-io*)
          (si::*top-level-hook*
           #'(lambda ()
               (cm::load-cminit)
               (setf *package* (find-package :cm)))))
      (si:save-system (namestring pathname))))
  )


#+(and clisp (or dos os/2 win32))
(progn
  (defparameter *user-homedir-pathname* "C:\\")
  (handler-bind ((t #'(lambda (c)
                        (when (find-restart 'continue)
                          (invoke-restart 'continue)))))
    (defun user-homedir-pathname () (pathname *user-homedir-pathname*)))

  )

#+clisp
(progn
  
  (defun address-of (object) (sys::address-of object))  

  (defmacro without-interrupts (&body body) `(progn ,@body))
  
  (defun shell (string) (lisp:shell string))

  (defun image-directory () 
    (and (boundp 'lisp::*image-pathname*) 
         lisp::*image-pathname*))
  
  (defun save-cm (pathname)
    (lisp:saveinitmem (namestring pathname) :quiet t
                      :init-function
                      #'(lambda ()
                          (load-cminit)
                          (setf *package* (find-package :cm))
                          (setf *readtable* *common-music-readtable*))))
  )

#+cmu
(progn
  
  (defun address-of (thing) (sys:%primitive c:make-fixnum thing))  

  (defun shell (string) 
    (extensions:run-program "/bin/csh" (list "-fc" string) :output t))

  (defmacro without-interrupts (&body body)
    `(system:without-interrupts ,@body))

  (defun image-directory ()
    (unix-image-directory 
     (first extensions:*command-line-strings*)
     (cdr (assoc ':path extensions:*environment-list*))))
 
  (defun save-cm (pathname)
    (extensions:save-lisp (namestring pathname)
                          :init-function
                          #'(lambda ()
                              (setf *readtable* *common-music-readtable*)
                              (setf *package* (find-package :cm))
                              (cm::load-cminit)
                              (lisp::%top-level))))
  )

#+aclpc
(progn
  
  (defun address-of (thing) NIL)  

  (defun shell (string) (warn "Shell not implemented in ACL/PC: ~S" string))
  
  (defmacro without-interrupts (&body body) `(progn ,@body))
  
  (defun image-directory () nil)
  
  (defun save-cm (pathname)
    (setf allegro:*session-init-fns*
          (append allegro:*session-init-fns*
                  (list #'(lambda ()
                            (setf *readtable* *common-music-readtable*)
                            (setf *package* (find-package :cm))))))
    (allegro:save-image :image-file (namestring pathname)))
  )

#+lispworks
(progn

  (defun shell (string) (foreign:call-system string "/bin/csh"))
  
  (defmacro without-interrupts (&body body)
    `(mp:without-interrupts ,@body))

  (defun image-directory ()
    (unix-image-directory (car sys::*line-arguments-list*)
                          (system::getenv "PATH")))

  (defun save-cm (pathname)
    (lispworks:save-image (namestring pathname) 
                          :gc t     ; Garbage collect
                          :type :user ; Clear global variables
                          :normal-gc t     ; normal-gc
                          :restart-function 
                          #'(lambda ()
                              (setf *package* (find-package package))
                              (setf *readtable* *common-music-readtable*)
                              (cm::load-cminit))))
  )

