//
//
// File:         pmcl-midi.hh 
//
// Summary:      Header file for the pmclmidi shared-library reimplementation
//               of the Apple MidiManager interface for the PPC version of
//               Macintosh Common Lisp (3.9 or higher).
//
// Dependencies: pmcl-midi.cc
//               pmcl-midi.lisp
//               pmcl-entry.lisp
//
//               ***********************************************************
//               Copyright (c) 1996 by Tobias Kunze and Rick Taube.  All
//               rights reserved.  Use and copying of this software and
//               preparation of derivative works based upon this software
//               are permitted and may be copied as long as no fees or
//               compensation are charged for use, copying, or accessing
//               this software and all copies of this software include this
//               copyright notice.  Suggestions, comments and bug reports
//               are welcome.  Please address email regarding this midi
//               implementation to: t@kunze.stanford.edu or
//               hkt@ccrma.stanford.edu.
//               ***********************************************************  
//
// Author:       Tobias Kunze
// E-Mail:       t@kunze.stanford.edu
// Org:          CCRMA, Stanford University
//
// Orig-Date:    18-May-96 at 11:43:48
// Last-Mod:     26-Jul-96 at 00:14:53 by Tobias Kunze
//
// Description:  Since the MidiManger's queues have a limited size of 0xffff
//               bytes, I maximize the number of schedulable events by
//               writing them out to myself through an invisible output port 
//               which is connected to a invisible input port's input queue.
//               Since we have 6 `standard', visible ports (time, inputA,
//               inputB, outputA, outputB and an invisible output port), the
//               maximum number of invisible input ports is 10 (we can't
//               have more than 16 MidiManager ports).  Invisible ports are 
//               allocated dynamically.
// 
//
//    MidiManager       Shared Library           Lisp
//    -------------------------------------------------------
//        
//                      |                        |
//         Time        <->  (unused)             |
//                      |                        |
//                      |    MidiReadEvent() -- <-  - ()
//                      |        :               |
//        Input         -> -.    v               |
//                      |    ReadHook() -------  -> - (midi-read-hook)
//       Input2         -> -'                    |
//                      |                        |
//     [Output2]       <-  -.                    |
//                      |    :                   |
//      Output1        <-  -.:                   |
//                      |    :                   |
//                      |    InvReadHook()       |
//                      |    :                   |
//    InvInput1     .-  -> -':                   |
//                 :    |    :                   |
//       ...       :    |    :                   |
//                 :    |    :                   |
//   [InvInputn]   :.-  -> -'                    |
//                 :    |                        |
//    InvOutput     `- <-  -------------------  <-  - (midi-write-event)
//                      |                        |
//
//
//               Only one InvInput port is connected to the InvOutput port at
//               any given time and termed the current invisible input port.
//               Since MidiManager doesn't provide a routine to check for
//               empty queues, invisible input ports keep track of the most
//               future event written to it and get deallocated by a periodic
//               MidiWakeUp call if this time is less than the current client
//               time.  This (MarkObsInPorts()) actually only marks the
//               port's reference number as obsolete and leaves it to the
//               LISP side to call the deallocation routine RemoveObsPorts().
//               
//               Flushing is similarly complicated: FlushQueues() sets the 
//               InvReadHook() to flushing mode and schedules another 
//               MidiWakeUp call to reset the read hook back to normal mode.
//               The flushing duration is a semi-global parameter, which is
//               set during initializationo by MidiCSetUp().
//
//
//#ident "$Revision: 1.0 $"
//
// Changes:
//    18-Jun-96  tk   moved MidiWriteEvent back to pmcl-midi.lisp
//    07-Jul-96  tk   Revision 1.0


#ifndef _PMCL_MIDI_H_
#define _PMCL_MIDI_H_

#include <midi.h>
#include <memory.h>

#ifndef TRUE
static const int TRUE = 1;
#endif
#ifndef FALSE
static const int FALSE = 0;
#endif

#ifndef Boolean
typedef unsigned char Boolean;
#endif

static const int kMaxInvInPorts = 10;
static const int kMidiSuccess = 0;
static const int kNoPort = 0;

class MidiDrConfig;             // forward declaration



// -------------------------------------------------------------------------
// 
// Dynamic Ports Class.
//

class DynamicPorts {
  friend MidiDrConfig;
  
#pragma export on
  // LISP Entry Points

  // Add a new dynamic port at position 0.  A return value of -1 indicates
  // that we can't allocate more dynamic ports
  friend int RegisterInvInPort(short refnum);
  // Cycle fInvInPorts so that index position is at position 0.
  // (We always write to position 0)
  friend void RotateInvInPorts(int places);
  // Deallocate obsolete ports.
  friend void RemoveObsPorts();
  // Return the adress of bOutputInProgress to LISP.
  friend unsigned char* GetbOutputInProgress();
#pragma export off

  // Toolbox Callbacks

  // This is a TimeProc scheduled by the MidiManager.  It looks for expired
  // inports at regular intervals, moving them from the fInvInPorts array to
  // fObsInPorts.  Actual deallocation is handled by RemoveObsPorts()
  friend extern pascal void MarkObsInPorts(long curTime, long myRefCon);

public:
  // Constructor
  DynamicPorts();
  // Deferred Initialization Function
  void Init(long bufidletimeout, long wakeupperiod);
  // Flush dynamic ports
  void FlushInvInPorts();

  // Yes, we have public data members:

  // This flag is set by %mm-write-message and %mm-all-notes-off to prevent 
  // MarkObsInPorts() from changing the status of ports we might be currently
  // outputting to from LISP
  Boolean bOutputInProgress;
  // Valid port reference numbers.  It is important this array is always
  // terminated by a zero int which serves as a `fill-pointer' to LISP, hence
  // the array is greater by one than kMaxInvInports
  short fInvInPorts[kMaxInvInPorts+1];

private:
  int fNumInvInPorts; 
  int fFirstFreeInObsInPorts;
  short fObsInPorts[kMaxInvInPorts]; // queue buffers marked obsolete

  long fBufIdleTimeOut;
  long fWakeUpPeriod;
};



// -------------------------------------------------------------------------
// 
// Midi Driver Configuration Class.
//


class MidiDrConfig {

#pragma export on
  // LISP Entry Points

  // Read pending events from the input queues.  This polls the queues which
  // then in turn call our ReadHook()
  friend int MidiReadEvent();
  // Set the read hook of invisible input ports to InvReadHook()
  friend void SetReadHook(short refNum);
  // Flush all pending output.
  friend void FlushQueues();
  // Initialization.
  friend short* MidiCSetUp(short timeport, short inportA, short inportB, 
                           short outportA, short outportB, short invoutport,
                           void (*lispreadhook) (long, long),
                           long wakeupperiod, long bufidletimeout, 
                           long flushduration);
#pragma export off

  // Toolbox Callbacks

  // The Input ReadHook.  This is called by the Midi Manager in response to
  // a MIDIPoll call from Lisp.  RefCon is unused.
  friend extern pascal short ReadHook(MIDIPacketPtr packet, long RefCon);
  // This readhook passes midi messages from a queue buffer to the actual
  // midi output ports.  RefCon is unused.
  friend extern pascal short InvReadHook(MIDIPacketPtr packet, long RefCon);
  // Terminates flushing mode after fFlushDuration milliseconds.
  friend extern pascal void EndFlushing(long curTime, long RefCon);

public:
  // Constructor
  MidiDrConfig(DynamicPorts* ports);

private:
  // Private Member Funcions
  inline int IsChannelMessage(MIDIPacketPtr packet);
  inline int WriteEncodedPacket(MIDIPacketPtr packet);

  // Data Members
  Boolean bFlushing;
  long fFlushDuration;
  void (* pLispReadHook) (long,long);

  // Our Ports
  int fTimePort;
  int fInPortA;                 // input channels 1-16
  int fInPortB;                 // input channels 17-32
  int fOutPortA;                // output channels 1-16
  int fOutPortB;                // output chnanels 17-32
  int fInvOutPort;              // the only port we write to


  // the variable ports
  DynamicPorts* pDynPorts;
};

#endif // _PMCL_MIDI_H_

//
// -*- EOF -*-
